//
// Copyright (c) 2009-2011, ERICSSON AB
// All rights reserved.
// 
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
// 
//     * Redistributions of source code must retain the above copyright notice, 
//       this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the ERICSSON AB nor the names of its contributors
//       may be used to endorse or promote products derived from this software
//       without specific prior written permission.
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
// OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//

#ifndef FELDSPAR_ARRAY_H
#define FELDSPAR_ARRAY_H

#include <stdint.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>

// TODO qualify the names to avoid clashes with Haskell names

struct array
{
    void*    buffer;   /* pointer to the buffer of elements */
    int32_t  length;   /* number of elements in the array */
    int32_t  elemSize; /* size of elements in bytes; (- sizeof(struct array)) for nested arrays */
    uint32_t bytes;    /* The number of bytes the buffer can hold */
};

/* Indexing into an array: */
/* Result: element of type 'type' */
#define at(type,arr,idx) (((type*)((arr)->buffer))[idx])


/* Deep array copy */
static inline void copyArray(struct array *to, struct array *from)
{
    assert(to);
    assert(from);
    if( from->elemSize < 0 )
    {
        unsigned i;
        for( i = 0; i < from->length; ++i )
            copyArray( &at(struct array, to, i), &at(struct array, from, i) );
    }
    else
    {
        assert(to->buffer);
        assert(from->buffer);
        memcpy( to->buffer, from->buffer, to->length * to->elemSize );
    }
}

/* Deep array copy to a given position */
static inline void copyArrayPos(struct array *to, unsigned pos, struct array *from)
{
    assert(to);
    assert(from);
    if( from->elemSize < 0 )
    {
        unsigned i;
        for( i = 0; i < from->length; ++i )
            copyArray( &at(struct array, to, i + pos), &at(struct array, from, i) );
    }
    else
    {
        assert(to->buffer);
        assert(from->buffer);
        memcpy( (char*)(to->buffer) + pos * to->elemSize, from->buffer, to->length * to->elemSize );
    }
}


/* Deep array copy with a given length */
void copyArrayLen(struct array *to, struct array *from, int32_t len);

/* Array length */
static inline int32_t getLength(struct array *arr)
{
  assert(arr);
  return arr->length;
}

/* (Re)set array length */
static inline void setLength(struct array *arr, int32_t len)
{
  assert(arr);
  assert(arr->bytes >= arr->length * abs(arr->elemSize));
  arr->length = len;
}

/* Reset array length by increasing it */
void increaseLength(struct array *arr, int32_t len);

#endif
